"""
Run test with:

    tox -c tox-full.ini -e py37-ubuntu_18_04_64 -- -s burlap/tests/test_apache.py::ApacheTests::test_diff

"""
from __future__ import print_function

import os
import shutil
import getpass

from fabric.api import settings

import burlap
from burlap.common import env, Satchel
from burlap.tests.base import TestCase
from burlap.project import project
from burlap.context import set_cwd
from burlap.deploy import STORAGE_LOCAL


class _TestSatchel(Satchel):

    name = 'test'

    def configure(self):
        pass


class ApacheTests(TestCase):

    def get_test_satchel(self):
        test = _TestSatchel()
        test.genv.hosts = ['localhost']
        test.genv.host_string = test.genv.hosts[0]
        return test

    def test_diff(self):
        """
        Confirm on a multi-site multi-host environment, apache correctly reports change.
        """

        test = self.get_test_satchel()

        print('Setting paths...')
        #env.plan_storage = STORAGE_LOCAL
        env.disable_known_hosts = True
        burlap_dir = os.path.abspath(os.path.split(burlap.__file__)[0])

        print('Initializing tmp directory...')
        d = '/tmp/test_apache_change'
        if os.path.isdir(d):
            shutil.rmtree(d)
        os.makedirs(d)

        activate_cmd = '. {d}/.env/bin/activate;'.format(d=d)
        with set_cwd(d):
            print('Creating project skeleton...')
            project.create_skeleton(
                project_name='test_apache_change',
                roles='prod',
                components='apache',
            )

            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)

            # Simulate multiple remote hosts by creating aliases of localhost.
            # Note, for testing this on your localhost for a user without passwordless sudo,
            # you may have to run: `sudo chmod 777 /etc/hosts`
            # This won't work on Travis, where these will instead be set in .travis.yml.
            print('Modifying /etc/hosts...')
            env.host_string = 'localhost'
            env.hosts = [env.host_string]
            env.user = getpass.getuser()
            hosts_updated = False
            with settings(warn_only=True):
                output = test.run('cat /etc/hosts')
                if 'test-dj-migrate-1' not in output:
                    for use_sudo in (False, True):
                        print('Attempting to append to /etc/hosts with use_sudo=%s...' % use_sudo)
                        #ret = append(filename='/etc/hosts', text='127.0.0.1 test-dj-migrate-1\n127.0.0.1 test-dj-migrate-2', use_sudo=use_sudo)
                        ret = test.append(filename='/etc/hosts', text='127.0.0.1 test-dj-migrate-1\n127.0.0.1 test-dj-migrate-2', use_sudo=use_sudo)
                        print('ret:', ret)
                        print('Checking /etc/hosts content...')
                        output = test.run('cat /etc/hosts')
                        print('output:', output)
                        if 'test-dj-migrate-1' in output:
                            hosts_updated = True
                            print('Success!')
                            break
                else:
                    hosts_updated = True
            assert hosts_updated

            os.system('ln -s %s %s/' % (burlap_dir, d))

            project.update_settings({
                    'plan_storage': STORAGE_LOCAL,
                    'plan_data_dir': os.path.join(d, 'plans'),
                    'services': ['apache'],
                    'default_site': 'testsite1',
                    'default_role': 'prod',
                    # This is necessary to stop get_current_hostname() from attempting to lookup our actual hostname.
                    '_ip_to_hostname': {
                        'test-dj-migrate-1': 'test-dj-migrate-1',
                        'test-dj-migrate-2': 'test-dj-migrate-2',
                    },
                    'apache_application_name': 'testsite1',
                    'apache_server_admin_email': 'sysadmin@mydomain.com',
                    'apache_server_aliases_template': '{apache_locale}.mydomain.com',
                    'apache_wsgi_dir_template': '/usr/local/{apache_application_name}/wsgi',
                    'apache_wsgi_processes': 1,
                    'apache_wsgi_threads': 0,
                },
                role='all')

            project.update_settings({
                    'hosts': ['test-dj-migrate-1', 'test-dj-migrate-2'],
                    'available_sites_by_host':{
                        'test-dj-migrate-1': [
                            'testsite1',
                        ],
                        'test-dj-migrate-2': [
                            'testsite2',
                        ]
                    },
                    'sites': {
                        'testsite1': {
                            'apache_domain_template': 'testsite1.test-dj-migrate-1.com',
                        },
                        'testsite2': {
                            'apache_domain_template': 'testsite2.test-dj-migrate-2.com',
                        },
                    },
                },
                role='prod')

            # Run a deployment preview.
            kwargs = dict(
                activate_cmd=activate_cmd,
            )
            status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.preview'.format(**kwargs))
            print('output:\n%s' % output)
            assert not status
            # The deployment preview should include both hosts.
            assert "[test-dj-migrate-1] Executing task 'deploy.preview'" in output
            assert "[test-dj-migrate-2] Executing task 'deploy.preview'" in output
            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            assert not os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)

            status, output = self.getstatusoutput(
                '{activate_cmd} '
                'fab prod debug.set_satchel_value:apache,site,xyz '
                'debug.show_satchel_items:apache'.format(**kwargs))
            assert not status
            assert ' = xyz' in output

            # Fake a deployment.
            #status, output = self.getstatusoutput((
                #'{activate_cmd} '
                #'fab prod debug.set_satchel_value:apache,site,abc '
                #'deploy.fake:set_satchels="apache-site-abc"').format(**kwargs))
            #assert not status
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)

            ## Confirm apache now reports no changes needing deployment.
            #status, output = self.getstatusoutput('{activate_cmd} fab prod deploy.preview'.format(**kwargs))
            #assert not status
            #assert "[test-dj-migrate-1] Executing task 'deploy.preview'" in output
            #assert "[test-dj-migrate-2] Executing task 'deploy.preview'" in output
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d)
            #print(open('%s/plans/prod/000/thumbprints/test-dj-migrate-1' % d).read())
            #assert os.path.isfile('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d)
            #print(open('%s/plans/prod/000/thumbprints/test-dj-migrate-2' % d).read())



from __future__ import print_function

from burlap.common import set_dryrun, run_or_dryrun, sudo_or_dryrun
from burlap.files import file # pylint: disable=redefined-builtin
from burlap.tests.functional_tests.base import TestCase

is_file = file.is_file

class CommonTests(TestCase):

    def test_dryrun(self):

        set_dryrun(True)
        run_or_dryrun('touch ~/abc.txt')
        assert not is_file('~/abc.txt')

        set_dryrun(1)
        run_or_dryrun('touch ~/def.txt')
        assert not is_file('~/def.txt')

        set_dryrun(False)
        run_or_dryrun('touch ~/mno.txt')
        assert not is_file('~/mno.txt')
        run_or_dryrun('rm -f ~/mno.txt')

        set_dryrun(0)
        run_or_dryrun('touch ~/xyz.txt')
        assert not is_file('~/xyz.txt')
        run_or_dryrun('rm -f ~/xyz.txt')

    def test_sudo(self):

        all_users = run_or_dryrun('cut -d: -f1 /etc/passwd')
        print('all users:', all_users)

        ret = (sudo_or_dryrun('whoami') or '').split('\n')[-1]
        print('ret0:', ret)
        self.assertEqual(ret, 'root')

        if 'travis' in all_users:
            target_user = 'travis'
        else:
            target_user = 'www-data'
        ret = (sudo_or_dryrun('whoami', user=target_user) or '').split('\n')[-1]
        print('ret1:', ret)
        self.assertEqual(ret, target_user)



from __future__ import print_function

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task
from burlap.common import only_hostname

RSYNC = 'rsync'

#DEPRECATED: TODO: remove tarball functionality, and rename to CodeSatchel
class TarballSatchel(Satchel):

    name = 'tarball'

    def set_defaults(self):

        self.env.clean = 1

        self.env.gzip = 1

        self.env.method = RSYNC

        self.env.rsync_source_dir = 'src'

        self.env.rsync_source_dirs = [] # This overrides rsync_source_dir

        self.env.rsync_target_dir = None

        self.env.rsync_target_host = '{user}@{host_string}:'

        self.env.rsync_auth = '--rsh "ssh -t -o StrictHostKeyChecking=no -i {key_filename}"'

        self.env.rsync_command_template = (
            'rsync '
            '--recursive --verbose --perms --times --links '
            '--compress --copy-links {exclude_str} '
            '--delete --delete-before --force '
            '{rsync_auth} '
            '{rsync_source_dir} '
            '{rsync_target_host}{rsync_target_dir}'
        )

        self.env.exclusions = [
            '*_local.py',
            '*.pyc',
            '*.svn',
            '*.tar.gz',
            '*.log',
            'twistd.pid',
            '*.sqlite',
        ]

        self.env.dir = '.burlap/tarball_cache'

        self.env.extra_dirs = []

        self.env.perm_user = 'www-data'

        self.env.perm_group = 'www-data'

        self.env.perm_chmod = None

        self.env.set_permissions = True

    @property
    def timestamp(self):
        from burlap.common import get_last_modified_timestamp
        r = self.local_renderer
        fn = r.env.rsync_source_dir
        if self.verbose:
            print('tarball.fn:', fn)
        return get_last_modified_timestamp(fn, ignore=[_ for _ in r.env.exclusions if '/' not in _])

    @task
    def changed(self):
        lm = self.last_manifest
        last_timestamp = lm.timestamp
        current_timestamp = self.timestamp
        self.vprint('last_timestamp:', last_timestamp)
        self.vprint('current_timestamp:', current_timestamp)
        ret = last_timestamp == current_timestamp
        print('NO change' if ret else 'CHANGED!')
        return ret

    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        manifest = super(TarballSatchel, self).record_manifest()
        manifest['timestamp'] = self.timestamp
        return manifest

    @task
    def set_permissions(self):
        r = self.local_renderer
        if r.env.rsync_target_dir:
            if r.env.perm_chmod:
                r.sudo('chmod -R {perm_chmod} {rsync_target_dir}')
            r.sudo('chown -R {perm_user}:{perm_group} {rsync_target_dir}')

    def _run_rsync(self, src, dst):
        print('rsync %s -> %s' % (src, dst))
        r = self.local_renderer
        r.env.hostname = only_hostname(r.genv.host_string)
        real_rsync_target_dir = r.env.rsync_target_dir
        try:
            # Rsync to a temporary directory where we'll have full permissions.
            tmp_dir = '/tmp/tmp_%s_%s' % (self.env.rsync_target_dir.replace('/', '_'), src.replace('/', '_'))
            r.env.rsync_target_dir = tmp_dir
            r.env.rsync_source_dir = src
            r.local(self.env.rsync_command_template)

            # Then rsync from the temp directory as sudo to complete the operation.
            r.env.rsync_source_dir = tmp_dir+'/*'
            r.env.rsync_target_dir = real_rsync_target_dir
            r.env.rsync_target_host = ''
            r.env.rsync_auth = ''
            r.sudo(self.env.rsync_command_template)
        finally:
            r.env.rsync_target_dir = real_rsync_target_dir

    @task
    def deploy_rsync(self, *args, **kwargs):
        r = self.local_renderer

        # Confirm source directories.
        src_dirs = list(self.env.rsync_source_dirs)
        if not src_dirs:
            src_dirs.append(self.env.rsync_source_dir)

        # Confirm target directories.
        assert self.env.rsync_target_dir

        r.env.exclude_str = ' '.join('--exclude=%s' % _ for _ in self.env.exclusions)

        for src_dir in src_dirs:
            self._run_rsync(src=src_dir, dst=self.env.rsync_target_dir)

        if self.env.set_permissions:
            self.set_permissions()

    @task(precursors=['gitchecker', 'packager', 'apache2', 'pip', 'user'])
    def configure(self, *args, **kwargs):
        if self.env.method == RSYNC:
            self.deploy_rsync(*args, **kwargs)

tarball_satchel = TarballSatchel()

deploy = tarball_satchel.configure



from __future__ import print_function

import os
import shutil
try:
    from commands import getoutput
except ImportError:
    from subprocess import getoutput

from burlap.constants import *
from burlap.common import get_satchel
from burlap.tests.base import TestCase
from burlap.context import set_cwd


class PipTests(TestCase):

    def test_pip_install(self):
        pip_satchel = get_satchel('pip')
        try:
            # Initialize tmp directory.
            d = '/tmp/test_pip_install'
            if os.path.isdir(d):
                shutil.rmtree(d)
            os.makedirs(d)

            # Install pip requirements.
            with set_cwd(d):

                # Create requirements file.
                os.makedirs('roles/all')
                with open('roles/all/pip-requirements.txt', 'w') as fout:
                    print('PyYAML\n', file=fout)

                # Install without the quiet flag
                pip_satchel.verbose = GLOBAL_VERBOSE
                pip_satchel.env.quiet_flag = ''
                pip_satchel.env.virtualenv_dir = '%s/.env' % d
                pip_satchel.configure()

                self.assertTrue(os.path.isdir(pip_satchel.env.virtualenv_dir))
                ret = getoutput('%s/bin/pip freeze | grep -i yaml' % pip_satchel.env.virtualenv_dir)
                print('pip freeze:\n', ret)
                self.assertTrue('PyYAML' in ret)

                # Delete the virtualenv.
                shutil.rmtree(os.path.join(d, '.env'))

                # Install with the quiet flag
                pip_satchel.verbose = GLOBAL_VERBOSE
                pip_satchel.env.quiet_flag = ' -q '
                pip_satchel.configure()

                self.assertTrue(os.path.isdir('.env'))
                ret = getoutput('.env/bin/pip freeze | grep -i yaml')
                print('pip freeze:\n', ret)
                self.assertTrue('PyYAML' in ret)

        finally:
            shutil.rmtree(d)
